<?php
class cmsCache {

    private static $instance;

    private $cacher;

    public $query_count = 0;


    public static function getInstance() {
        if (self::$instance === null) {
            self::$instance = new self;
        }
        return self::$instance;
    }

    public function __construct() {
        // >> <WebMan> - optim
        if (cmsConfig::get('cache_enabled')) {
            
            $cacher_class = 'cmsCache' . string_to_camel('_', cmsConfig::get('cache_method'));
            $this->cacher = new $cacher_class();

        } else {
            
            $this->cacher = false;
            
        }
        // << </WebMan> - optim
    }

    public function start() {
        // >> <WebMan> - optim
        if (!$this->cacher) { return false; }
        // << </WebMan> - optim
        $this->cacher->start();
    }

    public function stop() {
        // >> <WebMan> - optim
        if (!$this->cacher) { return false; }
        // << </WebMan> - optim
        $this->cacher->stop();
    }

    public function set($key, $value, $ttl=false){

        // >> <WebMan> - debug, optim
        if (!$this->cacher) { return false; }
        
        if (DEBUG_ON) {
          $debug = cmsDebug::getInstance();
          $debug->beginCache();
        }

        if (!$ttl) { $ttl = cmsConfig::get('cache_ttl'); }

        $ret = $this->cacher->set($key, $value, $ttl);
        if (DEBUG_ON) $debug->endCache("set", $key, $ret);
        // << </WebMan> - debug, optim
        
        return $ret;

    }

    public function has($key){

        // >> <WebMan> - optim
        if (!$this->cacher) { return false; }
        // << </WebMan> - optim

        return $this->cacher->has($key);

    }

    public function get($key){

        // >> <WebMan> - debug, optim
        if (!$this->cacher) { return false; }
        
        if (DEBUG_ON) {
          $debug = cmsDebug::getInstance();
          $debug->beginCache();
        }

        if (!$this->has($key)) {
          if (DEBUG_ON) $debug->endCache("get", $key, "Not in cache");
          return false;
        }
        
        $value = $this->cacher->get($key);

        if (cmsConfig::get('debug') && $value) {
            $this->query_count++;
        }
        
        if (DEBUG_ON) $debug->endCache("get", $key, (is_array($value) ? count($value, COUNT_RECURSIVE) : 0));
        // << </WebMan> - debug, optim

        return $value;

    }

    public function clean($key){

        // >> <WebMan> - optim
        if (!$this->cacher) { return false; }
        // << </WebMan> - optim

        return $this->cacher->clean($key);

    }

}
